#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <ctype.h>
#include <time.h>

#include <unistd.h>
#include <netdb.h>

#include <sys/stat.h>

#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <sys/fcntl.h>

#define	VERSION	"2.0"

static int	sock;
static int	throttle;
static int	maxblock	= BUFSIZ;
static int	alarm_seconds;
static int	alarm_count;
static int	alarm_timeout;
static int	alarm_signal;

static long	copycount;

static void	alarm_ticker(void);

#ifdef	NO_SSL
#include "md5.h"
#else
#include "rsa.h"       /* SSLeay stuff */
#include "crypto.h"
#include "x509.h"
#include "pem.h"
#include "ssl.h"
#include "err.h"

static SSL	*ssl;
static SSL_CTX	*ssl_ctx;
#endif

static void
ex(const char *s)
{
  alarm(0);
  alarm(0);
  perror(s);
  sleep(1);
  exit(2);
}

static void
exs(const char *s)
{
  alarm(0);
  alarm(0);
  perror(s);
  ERR_print_errors_fp(stdout);
  sleep(1);
  exit(2);
}

#ifndef	NO_SSL
static void
tino_ssl_close(void)
{
  SSL_free(ssl);

  SSL_CTX_free(ssl_ctx);
}

void
tino_ssl_client(void)
{
  SSL_load_error_strings();

  ssl_ctx = SSL_CTX_new(SSLv23_server_method());
  if (!ssl_ctx)
    ex("SSL_CTX_new");
  
  ssl = SSL_new(ssl_ctx);
  if (!ssl)
    exs("SSL_new");

  if (SSL_connect(ssl))
    exs("SSL_connect");
}

void
tino_ssl_server(char *cert)
{
  SSL_load_error_strings();

  ssl_ctx = SSL_CTX_new(SSLv23_server_method());
  if (!ssl_ctx)
    ex("SSL_CTX_new");
  
  if (SSL_CTX_use_RSAPrivateKey_file(ssl_ctx, cert, SSL_FILETYPE_PEM))
    exs("SSL_CTX_use_RSA...");
  
  if (SSL_CTX_use_certificate_file(ssl_ctx, cert, SSL_FILETYPE_PEM))
    exs("SSL_CTX_use_cert...");

  ssl = SSL_new(ssl_ctx);
  if (!ssl)
    exs("SSL_new");

  SSL_set_fd(ssl, sock);

  if (SSL_accept(ssl))
    exs("SSL_accept");
}
#endif

void *
re_alloc(void *buf, size_t len)
{
  if (!len)
    len	= 1;
  if ((buf=(buf ? realloc(buf, len) : malloc(len)))==NULL)
    ex("malloc");
  return buf;
}

#define	USE_SIGACTION

static void
alarmhandler(int sig)
{
#ifndef USE_SIGACTION
  if (signal(SIGALRM, alarmhandler))
    ex("signal");
#endif
  alarm(alarm_seconds);
  if (++alarm_signal>alarm_count)
    ex("alarm");
}

static void
alarm_init(int s, int c)
{
#ifdef USE_SIGACTION
  struct sigaction sa;

  memset(&sa, 0, sizeof sa);
  sa.sa_handler	= alarmhandler;

  if (sigaction(SIGALRM, &sa, NULL))
    ex("sigaction");
#else
  signal(SIGALRM, alarmhandler);
#endif

  alarm_seconds	= s;
  alarm_count	= c;
  alarm_timeout	= 0;
  alarm_signal	= 0;

  alarm(s);
}

static void
alarm_process(void)
{
  if (alarm_signal)
    {
      if (++alarm_timeout>alarm_count)
	ex("timeout");

      alarm_ticker();

      alarm_signal	= 0;
    }
}

static void
alarm_reset(void)
{
  alarm_timeout	= 0;
}

static void
writen(char *buf, size_t len, int fo)
{
  static unsigned long	delta;
  static clock_t	last;
  int			pos, put;

  alarm_reset();
  for (pos=0; pos<len; pos+=put)
    {
      size_t	max;

      max	= len-pos;
      if (max>maxblock)
	max	= maxblock;

#ifndef NO_SSL
      if (ssl && sock==fo)
	put	= SSL_write(ssl, buf+pos, max);
      else
#endif
	put	= write(fo, buf+pos, max);
      alarm_process();

      if (put>0)
	{
	  alarm_reset();
	  if (throttle)
	    {
	      clock_t	now;
	      long	hold;

	      /* This code is self synchronizing
	       * to any type of problem.
	       */
	      delta	+= put;
	      now	=  clock();
	      delta	-= ((now-last)*(throttle<<10))/CLOCKS_PER_SEC;
	      last	=  now;
	      if (delta<0)
		delta	=  0;
	      hold	=  delta*1000/throttle;
	      if (hold>10000)
		usleep(hold);
	    }
	  continue;
	}

      if (!put)
	ex("premature EOF on output");

      put	= 0;
      if (errno!=EINTR && errno!=EAGAIN)
	ex("output");
    }
}

static int
readn(int fd, void *buf, size_t max)
{
  alarm_reset();
  for (;;)
    {
      int	got;

#ifndef NO_SSL
      if (ssl && sock==fd)
	got	= SSL_read(ssl, buf, max);
      else
#endif
	got	= read(fd, buf, max);
      alarm_process();

      if (got>=0)
	{
	  alarm_reset();
	  return got;
	}

      if (errno!=EINTR && errno!=EAGAIN)
	ex("input");
    }
}

static void
docopy(int fi, int fo)
{
  char		*buf;
  int		got;
  size_t	bs;

  bs	= maxblock;
  if (bs<BUFSIZ)
    bs	= BUFSIZ;
  buf	= re_alloc(NULL, bs);
  while ((got=readn(fi, buf, bs))!=0)
    {
      writen(buf, got, fo);
      copycount	+= got;
    }
}

void
swrite(char *s)
{
  size_t	len;

  len	= strlen(s);

  writen(s, len, sock);
  writen("\n", 1, sock);
}

char *
sread(void)
{
  char		buf[1024];
  char		*out	= 0;
  size_t	len	= 0;
  int		got;

  while ((got=readn(sock, buf, sizeof buf))>0)
    {
      char	*tmp;

      out	= re_alloc(out, len+got+1);
      memcpy(out+len, buf, got);
      out[len+got]	= 0;
      if ((tmp=strchr(out+len, '\n'))!=0)
	{
	  *tmp	= 0;
	  return out;
	}
      len	+= got;
    }
  out		= re_alloc(out, len+1);
  out[len]	= 0;
  return out;
}

char *
md5sum(int fd, long max)
{
  MD5_CTX	ctx;
  char		buf[BUFSIZ], digest[16];
  int		got, i;
  char		*out, *ptr;
  long		total;

  MD5_Init(&ctx);
  total	= 0;
  while (total!=max && (got=readn(fd, buf, sizeof buf))>0)
    {
      if (max>=total && max<total+got)
	got	= max-total;
      MD5_Update(&ctx, buf, got);
      total	+= got;
    }
  MD5_Final(digest, &ctx);

  out	= re_alloc(NULL, 32+32);
  sprintf(out, "%ld ", total);
  ptr	= out+strlen(out);
  for (i=0; i<16; i++)
    {
      *ptr++	= "0123456789abcdef"[(digest[i]>>4)&0xf];
      *ptr++	= "0123456789abcdef"[(digest[i]>>0)&0xf];
    }
  *ptr	= 0;
  return out;
}
